/*
 * main.c
 *
 *  Created on: Apr 22, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "utilities.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "lambda.h"
#include "ProblemDimension.h"
#include "FileMapAndParameters.h"
#include "averagingKernel.h"
#include "errorHandler.h"
#include "RLSErrorMessages.h"
#include "InversionData.h"
#include "InputData.h"
#include "RLSInvert.h"
#include "Model.h"
#include "fftw.h"
#include "linearEqSolve.h"

#define DEBUG 0
#define DEBUG1 0

static char* TEXT_EXTENSION = ".txt";

BOOLEAN RLSDimensionsOk3dPrelim(MAP_DEPTH_MATRIX M, MAP_DEPTH_MATRIX d, MAP_DEPTH_MATRIX C, MATRIX R,
		int numberOfMaps) {
	char cmsg[MAX_ERROR_MSG_LENGTH];
	BOOLEAN problem = FALSE;
	printf("Preliminary check dimensions of data ...\n");
	printf(" (nX,nY,M,nZ,nV) where M = # of maps\n");
	printf("M(%1d,%1d,%1d,%1d,%1d)\n", M.nX, M.nY, numberOfMaps, M.nZ, M.nV);
	printf("C(%1d,%1d,%1d,%1d,%1d) should be C(%1d,%1d,%1d,%1d,%1d)\n", C.nX, C.nY, C.nT, C.nZ, C.nV,
			M.nX, M.nY, numberOfMaps, numberOfMaps, 1);
	printf("d(%1d,%1d,%1d,%1d,%1d)\n", d.nX, d.nY, numberOfMaps, d.nZ, d.nV);
	printf("R(%1d,%1d)\n", R.nRow, R.nCol);
	if ((M.nX != C.nX) && (M.nY != C.nY)) {
		problem = FALSE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "(M.x, M.y) is (%d,%d); (C.x, C.y) is (%d,%d)", M.nX, (M.nY-1)*2, C.nX, (C.nY-1)*2);
		reportError(KernelCovXYSame, cmsg, ERROR);
	}
	int dnRow = numberOfMaps; // d.nT;
	int MnRow = numberOfMaps; // M.nT;
	// check dimensions of CIKFull.
	if (dnRow != C.nZ) {
		problem = FALSE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "d.nRow is %d; C.nCol is %d", dnRow, C.nZ);
		reportError(DnNotEqualCInRow, cmsg, ERROR);
	}
	if (MnRow != C.nT) {
		problem = FALSE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nRow is %d; C.nRow is %d", MnRow, C.nT);
		reportError(MnRowNotEqualCInRow, cmsg, ERROR);
	}

	int MnCol;
	MnCol = M.nZ;


	// Check orientation of matrices and transpose as needed to fit
	// expected pattern
	// CI must be square
	if (C.nZ != C.nT) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "CI is %d by %d", C.nZ, C.nT);
		reportError(MatrixSquare, cmsg, ERROR);
	}
	// R must be square
	if (R.nCol != R.nRow) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "R is %d by %d", R.nRow, R.nCol);
		reportError(MatrixSquare, cmsg, ERROR);
	}

	// check dimensions of C.
	if (dnRow != C.nZ) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "d.nRow is %d; C.nCol is %d", dnRow, C.nZ);
		reportError(DnNotEqualCInRow, cmsg, ERROR);
	}
	if (MnRow != C.nT) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nRow is %d; C.nRow is %d", MnRow, C.nT);
		reportError(MnRowNotEqualCInRow, cmsg, ERROR);
	}

	//if ((MnCol != R.nRow) && (getProblemDimension() != PDC.D5)) {
	//	problem = TRUE;
	//	snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nCol is %d; R.nRow is %d", MnCol, R.nRow);
	//	reportError(MnColNotEqualRnRow, cmsg, ERROR);
	//}
	if (getProblemDimension() == PDC.D5) {
		MnCol = M.nZ * M.nV;
	}
	if (MnCol != R.nRow) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nCol is %d; R.nRow is %d", MnCol, R.nRow);
		reportError(MnColNotEqualRnRow, cmsg, ERROR);
	}

	return problem;
}

BOOLEAN RLSDimensionsOk3d(MAP_DEPTH_MATRIX M, MAP_DEPTH_MATRIX d, MAP_DEPTH_MATRIX C, MATRIX R) {
	char cmsg[MAX_ERROR_MSG_LENGTH];
	BOOLEAN problem = FALSE;
	printf("Verifying dimensions of full data ...\n");
	printf(" (nX,nY,M,nZ,nV) where M = # of maps\n");
	printf("M(%1d,%1d,%1d,%1d,%1d)\n", M.nX, M.nY, M.nT, M.nZ, M.nV);
	printf("C(%1d,%1d,%1d,%1d,%1d) should be C(%1d,%1d,%1d,%1d,%1d)\n", C.nX, C.nY, C.nT, C.nZ, C.nV,
			M.nX, M.nY, M.nT, M.nT, 1);
	printf("d(%1d,%1d,%1d,%1d,%1d)\n", d.nX, d.nY, d.nT, d.nZ, d.nV);
	printf("R(%1d,%1d)\n", R.nRow, R.nCol);
	if ((M.nX != C.nX) && (M.nY != C.nY)) {
		problem = FALSE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "(M.x, M.y) is (%d,%d); (C.x, C.y) is (%d,%d)", M.nX, (M.nY-1)*2, C.nX, (C.nY-1)*2);
		reportError(KernelCovXYSame, cmsg, ERROR);
	}
	int dnRow = d.nT;
	int MnRow = M.nT;
	// check dimensions of CIKFull.
	if (dnRow != C.nZ) {
		problem = FALSE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "d.nRow is %d; C.nCol is %d", dnRow, C.nZ);
		reportError(DnNotEqualCInRow, cmsg, ERROR);
	}
	if (MnRow != C.nT) {
		problem = FALSE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nRow is %d; C.nRow is %d", MnRow, C.nT);
		reportError(MnRowNotEqualCInRow, cmsg, ERROR);
	}

	int MnCol;
	MnCol = M.nZ;


	// Check orientation of matrices and transpose as needed to fit
	// expected pattern
	// CI must be square
	if (C.nZ != C.nT) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "CI is %d by %d", C.nZ, C.nT);
		reportError(MatrixSquare, cmsg, ERROR);
	}
	// R must be square
	if (R.nCol != R.nRow) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "R is %d by %d", R.nRow, R.nCol);
		reportError(MatrixSquare, cmsg, ERROR);
	}

	// check dimensions of C.
	if (dnRow != C.nZ) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "d.nRow is %d; C.nCol is %d", dnRow, C.nZ);
		reportError(DnNotEqualCInRow, cmsg, ERROR);
	}
	if (MnRow != C.nT) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nRow is %d; C.nRow is %d", MnRow, C.nT);
		reportError(MnRowNotEqualCInRow, cmsg, ERROR);
	}

	//if ((MnCol != R.nRow) && (getProblemDimension() != PDC.D5)) {
	//	problem = TRUE;
	//	snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nCol is %d; R.nRow is %d", MnCol, R.nRow);
	//	reportError(MnColNotEqualRnRow, cmsg, ERROR);
	//}
	if (getProblemDimension() == PDC.D5) {
		MnCol = M.nZ * M.nV;
	}
	if (MnCol != R.nRow) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nCol is %d; R.nRow is %d", MnCol, R.nRow);
		reportError(MnColNotEqualRnRow, cmsg, ERROR);
	}

	return problem;
}

BOOLEAN RLSDimensionsOk1d(MATRIX M, MATRIX d, MATRIX C, MATRIX R) {
	char cmsg[80];
	BOOLEAN problem = FALSE;
	int dnRow, MnRow, MnCol;

	dnRow = d.nRow;
	MnRow = M.nRow;
	MnCol = M.nCol;

	// Check orientation of matrices and transpose as needed to fit
	// expected pattern
	// CI must be square

	if (C.nCol != C.nRow) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "CI is %d by %d", C.nRow, C.nCol);
		reportError(MatrixSquare, cmsg, ERROR);
	}

	// R must be square
	if (R.nCol != R.nRow) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "R is %d by %d", R.nRow, R.nCol);
		reportError(MatrixSquare, cmsg, ERROR);
	}
	if (dnRow != C.nCol) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "d.nRow is %d; CI.nCol is %d", dnRow, C.nCol);
		reportError(DnNotEqualCInRow, cmsg, ERROR);
	}
	if (MnRow != C.nRow) {
		// try the transpose of M
		if (MnRow != C.nRow) {
			problem = TRUE;
			snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nRow is %d; CI.nRow is %d", MnRow, C.nRow);
			reportError(MnRowNotEqualCInRow, cmsg, ERROR);
		}
	}
	if (MnCol != R.nRow) {
		problem = TRUE;
		snprintf(cmsg, MAX_ERROR_MSG_LENGTH, "M.nCol is %d; R.nRow is %d", MnCol, R.nRow);
		reportError(MnColNotEqualRnRow, cmsg, ERROR);
	}
	return problem;
}

int determineStatusPoint(ikx) {
	if (ikx < 10)
		return 1;
	if (ikx < 200)
		return 10;
	if (ikx < 800)
		return 50;
	return 100;
}

void printStatus(double lambda1, double lambda2, int dim) {
	if (dim >=3) {
		printf("\n Lambda = %f", lambda1);
		printf("; Lambda2 = %f", lambda2);
		printf("\n");
	}
}

/**
 * Initialize inversion structure with data from input files and
 * command-line arguments.
 * @param[in] inputParms Pointer to input data files and inversion parameters.
 * @param[in] commandLineOutputDir Output directory specified on command line (optional).
 * @param[in] debug If running with debug option.
 * @returns InversionData structure with matrices filled with input data.
 */
InversionData *RLSInit3d(InputParms inputParms, char *commandLineOutputDir, BOOLEAN debug) {

	MATRIX R; // symmetric positive-definite matrix
	LambdaSet lambdaSet, lambdaSet2;
	MAP_DEPTH_MATRIX MFull;
	MAP_DEPTH_MATRIX dFull;
	MAP_DEPTH_MATRIX CIKFull;
	MAP_DEPTH_MATRIX CKFull;
	char *RFName;

	printf("Memory Usage at Initialization: %ld\n", getMemoryReport());

	InversionData *inversionData;

	inversionData = (InversionData *) td_malloc(sizeof(InversionData));
	inversionData->debug = debug;

	if (commandLineOutputDir == NULL) {
		inversionData->outputBaseDir = newBaseDirectory(
				inputParms.outputBaseDirName);
	} else {
		inversionData->outputBaseDir = newBaseDirectory(commandLineOutputDir);
	}

	RFName = inputParms.InputRegFileName;

	lambdaSet = inputParms.lambdaSet1;
	lambdaSet2 = inputParms.lambdaSet2;

	// read all the data from the files

	// 3d problem, read, transform and coalesce kernel files into giant matrix MFull
	int useFourierTransform = TRUE;
	printf("Reading first kernel file ...\n");
	MAP_DEPTH_MATRIX M1, d1;
	M1 = readMapDepthMatrixKernel(inputParms.d3.InputKernelBaseDirName,
			inputParms.d3.KernelFilenames, 1,
			useFourierTransform);
	// Read, transform and coalesce map files into giant matrix dFull
	printf("Reading first data file ...\n");
	d1 = readMapDepthMatrix(inputParms.d3.InputMapBaseDirName,
			inputParms.d3.MapFilenames, 1,
			useFourierTransform);


	// Read and transform noise covariance files into giant matrix CIKFull
	printf("Reading noise covariance files ...\n");
	// really, this is the noise covariance, not the inverse.  We'll calculate the inverse after
	// Fourier transform
	CKFull = readMapDepthMatrixFromOneFile(inputParms.InputICovarFileName,
			useFourierTransform);

	printf("Reading Regularization file: %s\n", RFName);
	R = toMatrix(readfits(RFName));
	MATRIX RP = transpose(R);
	if (DEBUG)
		print_matrix("RP", RP.nRow, RP.nCol, RP.data, RP.nCol, ROW_MAJOR);
	freeMatrix(R);
	R = RP;

	MATRIX temp = ConvertToComplex(R);
	freeMatrix(R);
	R = temp;

	// check dimensions of matrix before reading all kernel and map files
	BOOLEAN problem;
	problem = RLSDimensionsOk3dPrelim(M1, d1, CKFull, R, inputParms.d3.nKernelMapFiles);
	if (problem)
		reportError(IncompatibleMatrices, "", FATAL);

	destroyMapDepthMatrix(M1);
	destroyMapDepthMatrix(d1);
	//------------- now read all the kernel and map files
	// 3d problem, read, transform and coalesce kernel files into giant matrix MFull
	useFourierTransform = TRUE;
	printf("Reading Kernel files ...\n");
	MFull = readMapDepthMatrixKernel(inputParms.d3.InputKernelBaseDirName,
			inputParms.d3.KernelFilenames, inputParms.d3.nKernelMapFiles,
			useFourierTransform);
	TdmMultiplyScalarTakeConjugate(&MFull, (2.0 * M_PI) * (2.0 * M_PI));

	// Read, transform and coalesce map files into giant matrix dFull
	printf("Reading data files ...\n");
	dFull = readMapDepthMatrix(inputParms.d3.InputMapBaseDirName,
			inputParms.d3.MapFilenames, inputParms.d3.nKernelMapFiles,
			useFourierTransform);

	//--------------------- end read all kernel and map files

	// check dimensions of all matrices before calculating inverse
	problem = RLSDimensionsOk3d(MFull, dFull, CKFull, R);
	if (problem)
		reportError(IncompatibleMatrices, "", FATAL);

	// multiply by N^2/(2pi)^2 to correct for two Fourier transforms instead of one
	double factorForCI;
	factorForCI = ((double) CKFull.nX * (double) CKFull.nX) / ((2.0 * M_PI)
			* (2.0 * M_PI));
	TdmMultiplyScalar(&CKFull, factorForCI);

	CIKFull = inverseMapDepth(CKFull); // take the inverse of sub matrices: CI = inv[CK(kx,ky,:,:)]
	// if the matrix cannot be inverted, exit gracefully.
	if (CIKFull.data == NULL) {
		reportError("Cannot invert covariance matrix", "RLSInit", FATAL);
	}

	initializeAvgKernelCalculation(inputParms.d3.AvgKernelIndices, MFull.nZ);
	printf("After AvgKernel initialization\n");

	freeInputParms(inputParms);

	inversionData->InputMatrices = newInputData();
	InputData *id;
	id = inversionData->InputMatrices;
	inversionData->IntermediateMatrices = (TempData *) td_malloc(sizeof(TempData));
	TempData *tempd;
	tempd = inversionData->IntermediateMatrices;
	id->K = MFull;
	id->C = CKFull;
	id->M = dFull;
	id->R = R;
	tempd->CI = CIKFull;

	inversionData->parameter1 = inputParms.lambdaSet1;
	inversionData->parameter2 = inputParms.lambdaSet2;
	inversionData->nthStepParm1 = 1;
	inversionData->nthStepParm2 = 1;
	inversionData->currentValueParm1 = inputParms.lambdaSet1.min;
	inversionData->currentValueParm2 = inputParms.lambdaSet2.min;
	inversionData->outputModel = (Model *) td_malloc(sizeof(Model));
	initializeOutputModel(inversionData->outputModel, inversionData->InputMatrices->K);

	printf(
			"\nCalculating using Lambda Min = %f, Lambda Max = %f, N Lambda = %d\n",
			lambdaSet.min, lambdaSet.max, lambdaSet.n);
	 printf("		 Lambda2 Min = %f, Lambda2 Max = %f, N Lambda2 = %d\n",
			lambdaSet2.min, lambdaSet2.max, lambdaSet2.n);

	return (inversionData);
}

/**
 * Initialize inversion structure with data from input files and
 * command-line arguments.
 * @param[in] inputParms Pointer to input data files and inversion parameters.
 * @param[in] commandLineOutputDir Output directory specified on command line (optional).
 * @param[in] debug If running with debug option.
 * @returns InversionData structure with matrices filled with input data.
 */
InversionData *RLSInit1d(InputParms inputParms, char *commandLineOutputDir, BOOLEAN debug) {
	MATRIX M;
	MATRIX R; // symmetric positive-definite matrix
	MATRIX d;
	MATRIX C;
	LambdaSet lambdaSet, lambdaSet2;

	printf("Memory Usage at Initialization: %ld\n", getMemoryReport());

	char *MFName, *RFName, *dFName, *CIFName;

	InversionData *inversionData;

	inversionData = (InversionData *) td_malloc(sizeof(InversionData));
	inversionData->debug = debug;

	if (commandLineOutputDir == NULL) {
		inversionData->outputBaseDir = newBaseDirectory(
				inputParms.outputBaseDirName);
	} else {
		inversionData->outputBaseDir = newBaseDirectory(commandLineOutputDir);
	}

	MFName = inputParms.d1.InputKernelFileName;
	dFName = inputParms.d1.InputMapFileName;
	RFName = inputParms.InputRegFileName;
	CIFName = inputParms.InputICovarFileName;

	lambdaSet = inputParms.lambdaSet1;
	lambdaSet2 = inputParms.lambdaSet2;

	lambdaSet2.min = 0.0;
	lambdaSet2.max = 1.0;
	lambdaSet2.n = 1;

	// if any vector files are .txt, read them as ASCII files with one number per line, or
	// a comma separated list, or a space or tab delimited list.

	//* read all the data from the files

	printf("Reading Kernel file: %s\n", MFName);
	M = toMatrix(readfits(MFName));
	printf("Reading data file: %s\n", dFName);
	// if vector is an ASCII file, go ahead and accept it.
	if (endsWith(dFName, TEXT_EXTENSION)) {
		VECTOR dv = readText(dFName);
		d = VectorToMatrix_Column(dv);
	} else { // assume it is a .fits file
		MATRIX dT = toMatrix(readfits(dFName));
		d = transpose(dT);
		freeMatrix(dT);
	}
	MATRIX MP;

	if (M.data == NULL)
		reportError(NoInputData, "main", FATAL);
	MP = transpose(M);

	printf("Reading Noise Covariance file: %s\n", CIFName);
	C = toMatrix(readfits(CIFName));
	MATRIX CIP = transpose(C);

	freeMatrix(C);
	freeMatrix(M);
	M = MP;
	C = CIP;

	printf("Reading Regularization file: %s\n", RFName);
	R = toMatrix(readfits(RFName));
	MATRIX RP = transpose(R);
	if (DEBUG)
		print_matrix("RP", RP.nRow, RP.nCol, RP.data, RP.nCol, ROW_MAJOR);
	freeMatrix(R);
	R = RP;

	//* check dimensions
	BOOLEAN problem;
	problem = RLSDimensionsOk1d(M, d, C, R);
	if (problem)
		reportError(IncompatibleMatrices, "", FATAL);

	freeInputParms(inputParms);

	inversionData->InputMatrices = newInputData();
	InputData *id;
	id = inversionData->InputMatrices;
	inversionData->IntermediateMatrices = (TempData *) td_malloc(sizeof(TempData));

	id->K1d = M;
	id->C1d = C;
	id->M1d = d;
	id->R = R;

	inversionData->parameter1 = inputParms.lambdaSet1;
	inversionData->parameter2 = lambdaSet2;
	inversionData->nthStepParm1 = 1;
	inversionData->nthStepParm2 = 1;
	inversionData->currentValueParm1 = inputParms.lambdaSet1.min;
	inversionData->currentValueParm2 = lambdaSet2.min;

	inversionData->outputModel = (Model *) td_malloc(sizeof(Model));
	inversionData->outputModel->bVECTORS = newMatrix(inputParms.lambdaSet1.n, M.nCol);

	// take advantage of data that doesn't change with lambda.
	InputData *input;
	input = inversionData->InputMatrices;
	tdSetup1d(input->K1d, input->M1d, input->R, inverseD(input->C1d), input->C1d);

	printf(
			"\nCalculating using Lambda Min = %f, Lambda Max = %f, N Lambda = %d\n",
			lambdaSet.min, lambdaSet.max, lambdaSet.n);

	double lambda1 = inversionData->currentValueParm1;
	double lambda2 = inversionData->currentValueParm2;
	setParameters(inversionData->outputModel, lambda1, lambda2);
	BOOLEAN debugPosition = FALSE;
	int values[1];
	values[0] = inversionData->nthStepParm1;
	addParmsToFilename2(inversionData->outputBaseDir, values, 1, debugPosition);

	return (inversionData);
}

/*
 * Advance parameter values and indices one step.
 * @param[in,out] inversionData Pointer to structure holding inversion data.
 */
void RLSNextParameter3d(InversionData *inversionData) {

	// for each lambda step
	double lambda1, lambda2;

	inversionData->nthStepParm2 += 1;
	if (inversionData->nthStepParm2 > inversionData->parameter2.n) {
		inversionData->nthStepParm1 += 1;
		inversionData->currentValueParm1 =
				nextLambdaValue(inversionData->parameter1, inversionData->nthStepParm1-1);
		inversionData->nthStepParm2 = 1;
		inversionData->currentValueParm2 = inversionData->parameter2.min;

	} else {
		inversionData->currentValueParm2 =
				nextLambdaValue(inversionData->parameter2, inversionData->nthStepParm2-1);
	}
	lambda1 = inversionData->currentValueParm1;
	lambda2 = inversionData->currentValueParm2;
	setParameters(inversionData->outputModel, lambda1, lambda2);
}

/*
 * Advance parameter values and indices one step.
 * @param[in,out] inversionData Pointer to structure holding inversion data.
 */
void RLSNextParameter1d(InversionData *inversionData) {

	int lambda1, lambda2;

	inversionData->nthStepParm2 += 1;
	if (inversionData->nthStepParm2 > inversionData->parameter2.n) {
		inversionData->nthStepParm1 += 1;
		inversionData->currentValueParm1 =
				nextLambdaValue(inversionData->parameter1, inversionData->nthStepParm1-1);
		inversionData->nthStepParm2 = 1;
		inversionData->currentValueParm2 = inversionData->parameter2.min;

	} else {
		inversionData->currentValueParm2 =
				nextLambdaValue(inversionData->parameter2, inversionData->nthStepParm2-1);
	}
	lambda1 = inversionData->currentValueParm1;
	lambda2 = inversionData->currentValueParm2;
	setParameters(inversionData->outputModel, lambda1, lambda2);
	BOOLEAN debugPosition = FALSE;
	int values[1];
	values[0] = inversionData->nthStepParm1;
	addParmsToFilename2(inversionData->outputBaseDir, values, 1, debugPosition);
}

BOOLEAN RLSMoreParameterSteps(InversionData *inversionData) {
	return ((inversionData->nthStepParm1 <= inversionData->parameter1.n) &&
		(inversionData->nthStepParm2 <= inversionData->parameter2.n));
}

/**
 * Perform inversion for RLS 3D problems.
 * @param[in,out] inversionData Pointer to structure containing input data and
 * solution (output) data.
 */
void RLSInvert3d(InversionData *inversionData) {

	// prepare the output filenames and directory
	double lambda1 = inversionData->currentValueParm1;
	double lambda2 = inversionData->currentValueParm2;
	setParameters(inversionData->outputModel, lambda1, lambda2);
	BOOLEAN debugPosition = FALSE;
	int values[2];
	values[0] = inversionData->nthStepParm1;
	values[1] = inversionData->nthStepParm2;
	addParmsToFilename2(inversionData->outputBaseDir, values, 2, debugPosition);

	int ikx, iky;
	double kx, ky;
	MATRIX kernelk, mapk, CIk, Ck;
	BOOLEAN dumpDebug = FALSE;
	double lambda_squared;

	printStatus(lambda1, lambda2, getProblemDimension());

	InputData *input;
	input = inversionData->InputMatrices;

	ikx = input->K.nX; // MFull.nX;
	iky = input->K.nY; // MFull.nY;

	int multiple;
	multiple = determineStatusPoint(ikx);

	BaseDirectory *debugFilename;
	debugFilename = inversionData->outputBaseDir;

	int x = 0; // wavenumber or frequency for transformed data; positive values 0..N/2-1; negative values N/2..N-1;
	// N = ikx
	int ii, jj; // indices for kx and ky respectively
	for (ii = 0; ii < ikx; ii++) {

		kx = ((2.0 * M_PI) / ikx) * x;
		// output some progress information
		if (((ii % multiple) == 0) || (ii == ikx - 1)) {
			int percent = (float) ii / (float) (ikx - 1) * 100.0;
			printf(
					" kx index = %3d out of %3d; \tx = %3d; \tkx = %10lf; %3d%% done",
					ii, ikx - 1, x, kx, percent);
			printf("\n");

		}

		for (jj = 0; jj < iky; jj++) {
			// extract the d vector and M matrix from dFull and MFull
			/// HERE -- extract and construct M = [M_x, M_y, M_z]
			kernelk = extractMapDepth(ii, jj, input->K); // MFull);
			mapk = extractMapDepth(ii, jj, input->M); // dFull);
			CIk = extractMapDepth(ii, jj, inversionData->IntermediateMatrices->CI);
			Ck = extractMapDepth(ii, jj, input->C);
			if (((ii < 3) && (jj < 3)) && inversionData->debug) {
				dumpDebug = TRUE;
				BOOLEAN debugPosition = TRUE;
				BOOLEAN compress = FALSE;
				addParmsToFilename(debugFilename, ii, jj, debugPosition);
				print_matrix2toFile(finishFilename(debugFilename, "CIk", compress, debugPosition), CIk, ROW_MAJOR);
				print_matrix2toFile(finishFilename(debugFilename, "kernelk", compress, debugPosition), kernelk, ROW_MAJOR);
				print_matrix2toFile(finishFilename(debugFilename, "mapk", compress, debugPosition), mapk, ROW_MAJOR);
				print_matrix2toFile(finishFilename(debugFilename, "Ck", compress, debugPosition), Ck, ROW_MAJOR);
				print_matrix2toFile(finishFilename(debugFilename, "R", compress, debugPosition), input->R, ROW_MAJOR);
			} else {
				dumpDebug = FALSE;
			}
			tdSetup3d(kernelk, mapk, input->R, CIk, Ck);

			// calculate lambda
			ky = ((2.0 * M_PI) / ikx) * jj; // divide by ikx to keep the grid spacing the same in both x & y
			double k_norm_squared = kx * kx + ky * ky;
			lambda_squared = lambda1 * lambda1 + lambda2 * lambda2
					* k_norm_squared;

			tdinvert3dFinish(lambda_squared, inversionData->IntermediateMatrices, dumpDebug, debugFilename);

			TempData *tempd;
			tempd = inversionData->IntermediateMatrices;
			insertMapDepth(ii, jj, tempd->m, &(inversionData->outputModel->bFull));
			freeMatrix(tempd->m);
			if (calcAvgKernelCov()) {
				insertMapDepth(ii, jj, tempd->A,
						&(inversionData->outputModel->AvgKern));
				freeMatrix(tempd->A);
				insertMapDepth(ii, jj, tempd->Cov,
						&(inversionData->outputModel->noise));
				freeMatrix(tempd->Cov);
			}
			insertMapDepth(ii, jj, tempd->dTilda, &(inversionData->outputModel->dTilde));
			freeMatrix(tempd->dTilda);
			saveGridValues(inversionData->outputModel, ii, jj, kx, ky, lambda_squared);
			tdCleanup();
		} // end for jj

		// update x; if ii+1 is past Nyquist frequency (N/2) then, we are into negative values; reset x to -N/2
		// otherwise, increment x to next wavenumber.
		if (((ii + 1) - ikx / 2) % ikx == 0)
			x = -ikx / 2;
		else
			x += 1;

	} // end for ii

}

/**
 * Perform inversion for RLS 1D problems.
 * @param[in,out] inversionData Pointer to structure containing input data and
 * solution (output) data.
 */
void RLSInvert1d(InversionData *inversionData) {

	BOOLEAN dumpDebug;
	int ii, jj; // indices for kx and ky respectively
	ii = 0;
	jj = 0;
	double lambda_squared;

	if (((ii < 3) && (jj < 3)) && inversionData->debug) {
		dumpDebug = TRUE;
		BOOLEAN debugPosition = TRUE;
		addParmsToFilename(inversionData->outputBaseDir, ii, jj, debugPosition);
	} else {
		dumpDebug = FALSE;
	}

	double lambda1 = inversionData->currentValueParm1;
	lambda_squared = lambda1 * lambda1;
	printStatus(lambda1, 0.0, getProblemDimension());

	tdinvert1dFinish(lambda_squared, inversionData->nthStepParm1, inversionData->outputModel, dumpDebug, inversionData->outputBaseDir);
}

/**
 * Write solution matrices to FITS files for RLS 3D inversion.
 * @param[in,out] inversionData pointer to solution data.
 */
void RLSComplete3d(InversionData *inversionData) {
	writeBVectors3d(inversionData->outputModel, inversionData->outputBaseDir);
	BOOLEAN useFT = TRUE;
	writeDTildaValues3d(inversionData->outputModel, inversionData->outputBaseDir, useFT);
	writeAvgKernValues3d(inversionData->outputModel, inversionData->outputBaseDir);
	writeGridValues3d(inversionData->outputModel, inversionData->outputBaseDir);
}

/**
 * Write solution matrices to FITS files for RLS 1D inversion
 * @param[in,out] inversionData pointer to solution data.
 */
void RLSComplete1d(InversionData *inversionData) {
	double lambda;
	double lambda_squared;

	lambda = inversionData->currentValueParm1;
	lambda_squared = lambda * lambda;
	writeAVectors1d(inversionData->outputModel, inversionData->outputBaseDir, lambda_squared);
	writeCovVectors1d(inversionData->outputModel, inversionData->outputBaseDir, lambda_squared);
}

/**
 * Free memory for RLS 3D
 * @param[in,out] inversionData pointer to data
 */
void RLSDestroy3d(InversionData *inversionData) {

	InputData *im;
	im = inversionData->InputMatrices;
	destroyMapDepthMatrix(im->K);
	destroyMapDepthMatrix(im->M);
	destroyMapDepthMatrix(im->C);
	freeMatrix(im->R);

	linearEqSolveCleanup();

	Model *model = inversionData->outputModel;
	if (calcAvgKernelCov()) {
		destroyMapDepthMatrix(model->AvgKern);
		destroyMapDepthMatrix(model->noise);
	}

	destroyMapDepthMatrix(model->bFull);
	destroyMapDepthMatrix(model->dTilde);
	freeMatrix(model->kxGridValues);
	freeMatrix(model->kyGridValues);
	freeMatrix(model->lambdaGridValues);
	td_free(model, sizeof(Model));

	TempData *tempd;
	tempd = inversionData->IntermediateMatrices;
	destroyMapDepthMatrix(tempd->CI);

	destroyBaseDirectory(inversionData->outputBaseDir);

	td_free(inversionData->InputMatrices, sizeof(InputData));
	td_free(tempd, sizeof(TempData));
	td_free(inversionData, sizeof(InversionData));

	if (calcAvgKernelCov()) {
		freeAvgKernelCalculation();
	}

	printf("Memory Usage  %ld \tFinal\n", getMemoryReport());

	exit(EXIT_SUCCESS);
}


/**
 * Free memory for RLS 1D
 * @param[in,out] inversionData pointer to data
 */
void RLSDestroy1d(InversionData *inversionData) {

	// since we are accumulating b-vectors into a matrix (each lambda step is a column in the b-matrix)
	// output b-vectors as a matrix
	writeBVectors1d(inversionData->outputModel, inversionData->outputBaseDir);

	tdCleanup();

	freeMatrix(inversionData->InputMatrices->M1d);
	freeMatrix(inversionData->InputMatrices->K1d);
	freeMatrix(inversionData->InputMatrices->C1d);
	freeMatrix(inversionData->InputMatrices->R);

	Model *model = inversionData->outputModel;
	td_free(model, sizeof(Model));

	destroyBaseDirectory(inversionData->outputBaseDir);

	td_free(inversionData->InputMatrices, sizeof(InputData));
	td_free(inversionData->IntermediateMatrices, sizeof(TempData));
	td_free(inversionData, sizeof(InversionData));

	linearEqSolveCleanup();

	printf("Memory Usage  %ld \tFinal\n", getMemoryReport());

	exit(EXIT_SUCCESS);
}

